import os
import sys
import time
import math
import yaml
import copy
import torch
import wandb
import traceback
from tqdm import tqdm
import torch.nn as nn
from torch.utils.data import DataLoader
from torch.amp import autocast, GradScaler
from contextlib import nullcontext
from .utils import *
from .models import initialize_model
from .schedulers import initialize_scheduler
from .datasets import initialize_dataset

def default_data_fn(data, target):
    return data, target

def train_epoch(epoch, train_loader, model, optimizer, loss_fn, acc_fn, data_fn=default_data_fn,
                scheduler=None, device="cpu", completed_steps=0, train_steps=None,
                output_dir="", model_name=None, val_loader=None,
                wandb_logging=True, wandb_metrics=["acc", "loss"],
                grad_clip_norm=None, accumulation_steps=0,
                mixed_precision=False,
                checkpoint_freq=None, val_freq=None, info_freq=None):
    # Default model name
    if model_name is None: model_name = model.__class__.__name__
    model.train()
    train_loss = 0
    train_acc = 0
    accumulated_batch_loss = 0
    accumulated_batch_acc = 0
    iterable = tqdm(train_loader, desc=f"Train Epoch {epoch}", leave=False, bar_format='{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}')
    scaler = GradScaler(device=device) if mixed_precision else None
    optimizer.zero_grad()
    for batch_idx, (data, target) in enumerate(iterable):
        with autocast(device_type=device) if mixed_precision else nullcontext():
            # Forward pass
            data = data.to(device)
            target = target.to(device)
            data, target = data_fn(data, target)
            output = model(data)
            
            # Accuracy
            with torch.no_grad():
                accuracy = acc_fn(output.clone(), target.clone())
                accumulated_batch_acc += accuracy
                train_acc += accuracy
            
            # Loss
            loss = loss_fn(output, target)
            if torch.isnan(loss).any():
                report_bad_params(model)
                raise ValueError(f"Loss is NaN: Epoch {epoch}, Batch {batch_idx}")
            if torch.isinf(loss).any():
                report_bad_params(model)
                raise ValueError(f"Loss is Inf: Epoch {epoch}, Batch {batch_idx}")
            batch_loss = loss.item()
            accumulated_batch_loss += batch_loss
            train_loss += loss.item()
        
        # Backward pass and gradient accumulation if applicable
        loss = loss / (accumulation_steps + 1)
        loss.backward() if not mixed_precision else scaler.scale(loss).backward()
        
        if (batch_idx + 1) % (accumulation_steps + 1) == 0:
            if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
            optimizer.step() if not mixed_precision else scaler.step(optimizer)
            optimizer.zero_grad()
            if scheduler: scheduler.step()
            if mixed_precision: scaler.update()
            
            # WandB logging
            log_lr = scheduler.get_last_lr()[0] if scheduler else optimizer.param_groups[0]["lr"]
            accumulated_batch_loss /= (accumulation_steps + 1)
            accumulated_batch_acc /= (accumulation_steps + 1)
            if wandb_logging:
                log_data = {}
                if "acc" in wandb_metrics: log_data["train/acc"] = accumulated_batch_acc
                if "loss" in wandb_metrics: log_data["train/loss"] = accumulated_batch_loss
                if "ppl" in wandb_metrics: log_data["train/ppl"] = math.exp(accumulated_batch_loss)
                if "lr" in wandb_metrics: log_data["misc/lr"] = log_lr
                if "seq_len" in wandb_metrics: log_data["misc/seq_len"] = train_loader.dataset.len
                wandb.log(log_data)
            accumulated_batch_loss = 0
            accumulated_batch_acc = 0
            
            # Post info
            if info_freq and completed_steps % info_freq == 0 and completed_steps > 0:
                tqdm.write(f'Train Epoch {epoch}: [{batch_idx}/{len(train_loader)}] LR: {log_lr:.1e}, Loss: {batch_loss:.4f}, Acc: {accuracy:.2f}%')
            
            # Checkpoint
            if checkpoint_freq and completed_steps % checkpoint_freq == 0 and completed_steps > 0:
                checkpoint(model_name, output_dir, model, optimizer, scheduler)
            
            # Validation
            if val_freq and completed_steps % val_freq == 0 and completed_steps > 0:
                if val_loader: val_epoch(model, val_loader, loss_fn, acc_fn, device=device, wandb_logging=wandb_logging, wandb_metrics=wandb_metrics)
                model.train()
            
            completed_steps += 1
        
        if train_steps is not None and completed_steps >= train_steps: break
    
    # Account for last accumulated batch
    if (batch_idx + 1) % (accumulation_steps + 1) != 0:
        completed_steps += 1
        if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
        optimizer.step() if not mixed_precision else scaler.step(optimizer)
        optimizer.zero_grad()
        if scheduler: scheduler.step()
        if mixed_precision: scaler.update()
        
        # WandB logging
        log_lr = scheduler.get_last_lr()[0] if scheduler else optimizer.param_groups[0]["lr"]
        accumulated_batch_loss /= (batch_idx % (accumulation_steps + 1)) + 1
        accumulated_batch_acc /= (batch_idx % (accumulation_steps + 1)) + 1
        if wandb_logging:
            log_data = {}
            if "acc" in wandb_metrics: log_data["train/acc"] = accumulated_batch_acc
            if "loss" in wandb_metrics: log_data["train/loss"] = accumulated_batch_loss
            if "ppl" in wandb_metrics: log_data["train/ppl"] = math.exp(accumulated_batch_loss)
            if "lr" in wandb_metrics: log_data["misc/lr"] = log_lr
            if "seq_len" in wandb_metrics: log_data["misc/seq_len"] = train_loader.dataset.len
            wandb.log(log_data)
    
    # Step sequence length if applicable
    if hasattr(train_loader.dataset, "step"):
        train_loader.dataset.step()
    
    train_loss /= len(train_loader)
    train_acc /= len(train_loader)
    
    return train_loss, train_acc, completed_steps

@ torch.no_grad()
def val_epoch(model, val_loader, loss_fn, acc_fn, data_fn=default_data_fn,
              device="cpu",
              wandb_logging=True, wandb_metrics=["acc", "loss"],):
    model.eval()
    val_loss = 0
    val_acc = 0
    start = time.time()
    iterable = val_loader
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        data, target = data_fn(data, target)
        output = model(data)
        val_loss += loss_fn(output, target).item()
        val_acc += acc_fn(output, target)

    total_time = time.time() - start
    val_loss /= len(val_loader)
    val_acc /= len(val_loader)
    
    tqdm.write(f'\033[93mVal Epoch: Loss: {val_loss:.4f}, Acc: {val_acc:.2f}%, Elapsed: {total_time:.3f}s\033[0m')
    if wandb_logging:
        log_data = {}
        if "acc" in wandb_metrics: log_data["val/acc"] = val_acc
        if "loss" in wandb_metrics: log_data["val/loss"] = val_loss
        if "ppl" in wandb_metrics: log_data["val/ppl"] = math.exp(val_loss)
        wandb.log(log_data)
    
    return val_loss, val_acc

@ torch.no_grad()
def test_epoch(model, test_loader, loss_fn, acc_fn, data_fn=default_data_fn,
               device="cpu",
               wandb_logging=True, wandb_metrics=["acc", "loss"],):
    model.eval()
    test_loss = 0
    test_acc = 0
    start = time.time()
    tqdm.write("")
    iterable = tqdm(test_loader, desc=f"Test Epoch", leave=False, bar_format='\033[92m{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}\033[0m')
    for data, target in iterable:
        data = data.to(device)
        target = target.to(device)
        data, target = data_fn(data, target)
        output = model(data)
        test_loss += loss_fn(output, target).item()
        test_acc += acc_fn(output, target)

    total_time = time.time() - start
    test_loss /= len(test_loader)
    test_acc /= len(test_loader)
    
    tqdm.write(f'\033[92mTest Epoch: Loss: {test_loss:.4f}, Acc: {test_acc:.2f}%, Elapsed: {total_time:.3f}s\033[0m\n')
    if wandb_logging:
        log_data = {}
        if "acc" in wandb_metrics: log_data["test/acc"] = test_acc
        if "loss" in wandb_metrics: log_data["test/loss"] = test_loss
        if "ppl" in wandb_metrics: log_data["test/ppl"] = math.exp(test_loss)
        wandb.log(log_data)
        
    return test_loss, test_acc

def train(epochs, train_steps, benchmark_name, model, train_loader, optimizer, loss_fn, acc_fn, data_fn=default_data_fn,
          scheduler=None, device="cpu",
          train_config=None, mixed_precision=False,
          output_dir="", model_name=None,
          val_loader=None, test_loader=None,
          local_log_path=None,
          wandb_logging=True, wandb_entity=None, wandb_project=None,
          wandb_metrics=["acc", "loss"],
          grad_clip_norm=None, accumulation_steps=0,
          checkpoint_freq=None, val_freq=None, info_freq=None):
    try:
        sys.stdout.write("\033[?25l")
        
        NoEcho.disable_echo()
        
        # Default model name
        if model_name is None: model_name = model.__class__.__name__
        
        print(f'\033[1m{benchmark_name} Benchmark\033[0m')
        print(f'\033[1m{model_name}\033[0m')
        print(f'\033[4mTotal params: {count_parameters(model):,}\033[0m\n')
        
        # WandB Initialization
        if wandb_logging:
            assert wandb_entity is not None, "WandB entity is required for logging."
            assert wandb_project is not None, "WandB project is required for logging."
            wandb.init(
                settings=wandb.Settings(silent=True),
                mode="online" if online() else "offline",
                entity=wandb_entity,
                project=wandb_project,
                name=model_name,
                config=train_config,
            )
        
        # Use multiple GPUs if available
        if torch.cuda.device_count() > 1: model = nn.DataParallel(model)
        
        # Allocate dynamic memory if applicable for dataset
        if hasattr(train_loader.dataset, "seq_len_range"):
            min_len, max_len = train_loader.dataset.seq_len_range()
            model = allocate_dynamic_memory(model, train_loader.batch_size, min_len, max_len, device)
        
        # Metrics
        train_losses = []
        train_accuracies = []
        if test_loader:
            test_losses = []
            test_accuracies = []
        else:
            test_losses = None
            test_accuracies = None
        
        # Train loop
        if epochs is not None:
            for epoch in range(1, epochs + 1):
                # Train epoch
                train_loss, train_acc, _ = train_epoch(
                    epoch=epoch, train_loader=train_loader, val_loader=val_loader,
                    model=model, optimizer=optimizer, scheduler=scheduler,
                    loss_fn=loss_fn, acc_fn=acc_fn, data_fn=data_fn, device=device,
                    output_dir=output_dir, model_name=model_name,
                    wandb_logging=wandb_logging, wandb_metrics=wandb_metrics,
                    grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                    mixed_precision=mixed_precision,
                    checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq
                )
                train_losses.append(train_loss)
                train_accuracies.append(train_acc)
                
                # Test epoch
                if test_loader:
                    test_loss, test_acc = test_epoch(
                        model, test_loader, loss_fn, acc_fn,
                        device=device,
                        wandb_logging=wandb_logging, wandb_metrics=wandb_metrics
                    )
                    test_losses.append(test_loss)
                    test_accuracies.append(test_acc)
                
                # Model checkpoint
                checkpoint(model_name=model_name, output_dir=output_dir, model=model, optimizer=optimizer, scheduler=scheduler)
        
        elif train_steps is not None:
            completed_steps = 0
            epoch = 1
            while completed_steps < train_steps:
                train_loss, train_acc, completed_steps = train_epoch(
                    epoch=epoch, completed_steps=completed_steps, train_steps=train_steps, train_loader=train_loader, val_loader=val_loader,
                    model=model, optimizer=optimizer, scheduler=scheduler,
                    loss_fn=loss_fn, acc_fn=acc_fn, data_fn=data_fn, device=device,
                    output_dir=output_dir, model_name=model_name,
                    wandb_logging=wandb_logging, wandb_metrics=wandb_metrics,
                    grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                    mixed_precision=mixed_precision,
                    checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq
                )
                train_losses.append(train_loss)
                train_accuracies.append(train_acc)
                
                # Test epoch
                if test_loader:
                    test_loss, test_acc = test_epoch(
                        model, test_loader, loss_fn, acc_fn,
                        device=device,
                        wandb_logging=wandb_logging, wandb_metrics=wandb_metrics
                    )
                    test_losses.append(test_loss)
                    test_accuracies.append(test_acc)
                
                # Model checkpoint
                checkpoint(model_name=model_name, output_dir=output_dir, model=model, optimizer=optimizer, scheduler=scheduler)
                
                epoch += 1
        
        # Final logging
        if local_log_path: log_info(log_path=local_log_path, model=model, model_name=model_name, configs=train_config, train_accuracies=train_accuracies, test_accuracies=test_accuracies)
        if wandb_logging: wandb.finish()
    
    except KeyboardInterrupt as e: raise e
    except Exception as e:
        if wandb_logging: wandb.finish()
        raise e
    finally:
        NoEcho.enable_echo()
        if wandb_logging: cleanup_wandb(wandb_entity, wandb_project)
        sys.stdout.write("\033[?25h")

def train_from_config_file(yaml_path, loss_fn, acc_fn, data_fn=default_data_fn, device="cpu"):
    """
    **Config file options:**
        global:
            benchmark_name: Name of the benchmark.
            output_dir: Directory to save model checkpoints.
            
            dataset:
                name: Class name of the dataset to use.
                splits: Dictionary of dataset splits with their configurations. (e.g., "train", "val", "test")
            
            logging (optional):
                info_freq (default: 100): Frequency of CLI logging training information. No CLI logging if unspecified.
                local_log_path: Path to save local logs.
                wandb: WandB logging configurations.
                    entity: WandB entity name.
                    project: WandB project name.
                    metrics (default: ["acc", "loss"]): List of metrics to log. (e.g., ["acc", "loss", "ppl", "lr", "seq_len"]).
            
            val_freq (default: 500): Frequency of validation during training. No validation if unspecified.
            checkpoint_freq (default: 500): Frequency of saving model checkpoints. No checkpointing if unspecified.
        
        experiments:
            **List item format:**
                general:
                    seed (default: 0): Random seed for reproducibility.
                    batch_size (default: 32): Batch size for training.
                    accumulation_steps (default: 0): Number of batches to accumulate gradients for.
                    train_steps (optional): Number of training steps. **(Mutually exclusive with epochs)**
                    epochs (optional): Number of epochs to train. **(Mutually exclusive with train_steps)**
                    grad_clip_norm (optional): Gradient clipping norm. No clipping if unspecified.
                    load_checkpoint (default: False): Whether to attempt loading model from checkpoint.
                    mixed_precision (default: False): Whether to use mixed precision for training.
                    num_workers (default: 0): Number of workers for data loading.
                
                model:
                    name: Model class name.
                    Model arguments...
                
                optimizer:
                    name: Optimizer class name (e.g., "SGD", "Adam", "AdamW").
                    exclude_weight_decay (default ["bias", "norm"]): List of parameter names to exclude from weight decay. Provide empty list to not exclude any parameters.
                    Optimizer arguments...
                
                scheduler (optional):
                    name: Scheduler class name (e.g., "ConstantLR", "LinearLR", "CosineAnnealingLR").
                    Scheduler arguments...
    
    Args:
        yaml_path (str): Path to YAML configuration file.
        loss_fn (Callable): A function to compute model loss. Args: (model_output, target). Returns: loss.
        acc_fn (Callable): A function to compute model accuracy. Args: (model_output, target). Returns: accuracy.
        data_fn (Callable, optional): A function to augment data and target before passing into model. Args: (data, target). Returns: (data, target).
        device (str, optional): Device to run training on. Defaults to cpu.
    """
    os.system('clear')
    
    with open(yaml_path, 'r') as f:
        configs = yaml.safe_load(f)
    
    # Extract global training configurations
    global_config = configs.get("global")
    benchmark_name = global_config.get("benchmark_name")
    output_dir = global_config.get("output_dir")
    
    # Initialize datasets
    dataset_config = global_config.get("dataset")
    dataset_name = dataset_config.get("name")
    dataset_splits = dataset_config.get("splits", {})
    if "train" in dataset_splits:
        dataset_args = dataset_splits["train"]
        train_dataset = initialize_dataset(dataset_name, **dataset_args)
    val_dataset = None
    if "val" in dataset_splits:
        dataset_args = dataset_splits["val"]
        val_dataset = initialize_dataset(dataset_name, **dataset_args)
    test_dataset = None
    if "test" in dataset_splits:
        dataset_args = dataset_splits["test"]
        test_dataset = initialize_dataset(dataset_name, **dataset_args)
    
    # Get logging configurations
    logging_config = global_config.get("logging", {})
    info_freq = logging_config.get("info_freq", 100)
    local_log_path = logging_config.get("local_log_path")
    wandb_config = logging_config.get("wandb", {})
    wandb_logging = wandb_config is not None
    wandb_entity = wandb_config.get("entity")
    wandb_project = wandb_config.get("project")
    wandb_metrics = wandb_config.get("metrics", ["acc", "loss"])
    
    # Get checkpointing configurations
    val_freq = global_config.get("val_freq", 500)
    checkpoint_freq = global_config.get("checkpoint_freq", 500)
    
    # Run all experiments
    successful_count = 0
    experiments = configs.get("experiments")
    for i, experiment in enumerate(experiments):
        print(f'\033[1mRunning Experiment [{i + 1}/{len(experiments)}]\033[0m\n')
        
        # Reset dataset sequence lengths if applicable
        if hasattr(train_dataset, "reset"): train_dataset.reset()
        if hasattr(val_dataset, "reset"): val_dataset.reset()
        if hasattr(test_dataset, "reset"): test_dataset.reset()
        
        general_config = copy.deepcopy(experiment.get("general"))
        general_config = try_to_float(general_config)
        
        # Set seed
        seed = general_config.get("seed", 0)
        torch.manual_seed(seed)
        if device == "cuda":
            torch.cuda.manual_seed_all(seed)
        
        mixed_precision = general_config.get("mixed_precision", False) and device=="cuda"
        
        # Initialize dataloaders
        batch_size = general_config.get("batch_size", 32)
        epochs = general_config.get("epochs", None)
        train_steps = general_config.get("train_steps", None)
        assert not (train_steps is None and epochs is None), "Either train_steps or epochs must be specified."
        assert not (train_steps is not None and epochs is not None), "Only one of train_steps or epochs can be specified."
        grad_clip_norm = general_config.get("grad_clip_norm", None)
        accumulation_steps = general_config.get("accumulation_steps", 0)
        num_workers = general_config.get("num_workers", 0)
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
        val_loader = None
        test_loader = None
        if val_dataset: val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
        if test_dataset: test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, persistent_workers=(num_workers>0), pin_memory=True, prefetch_factor=2 if num_workers > 0 else None)
        
        model_config = copy.deepcopy(experiment.get("model"))
        model_name = model_config.pop("name")
        model_config = try_to_float(model_config)
        model_args = model_config.copy()
        model_args["device"] = device
        
        # Initialize model
        model = initialize_model(model_name, **model_args)
        
        # Initialize optimizer
        def initialize_optimizer(name, *args, **kwargs):
            optimizer_class = getattr(sys.modules["torch.optim"], name, None)
            return optimizer_class(*args, **kwargs)
        optimizer_config = copy.deepcopy(experiment.get("optimizer"))
        optimizer_name = optimizer_config.pop("name")
        optimizer_config = try_to_float(optimizer_config)
        weight_decay = float(optimizer_config.get("weight_decay", 0.0))
        exclude_weight_decay = optimizer_config.pop("exclude_weight_decay", None)
        apply_weight_decay_args = dict(
            model=model,
            weight_decay=weight_decay,
        )
        if exclude_weight_decay is not None: apply_weight_decay_args["exclude"] = exclude_weight_decay
        optimizer_config["params"] = apply_weight_decay(**apply_weight_decay_args)
        optimizer = initialize_optimizer(optimizer_name, **optimizer_config)
        
        # Initialize scheduler if specified
        scheduler = None
        scheduler_config = copy.deepcopy(experiment.get("scheduler", {}))
        if scheduler_config:
            scheduler_name = scheduler_config.pop("name")
            scheduler_config = try_to_float(scheduler_config)
            scheduler_config["optimizer"] = optimizer
            scheduler = initialize_scheduler(scheduler_name, **scheduler_config)
        
        # Load model from checkpoint if specified
        load_from_checkpoint = general_config.get("load_checkpoint", False)
        if load_from_checkpoint:
            model, optimizer, scheduler = load_checkpoint(
                model_name=model_name, output_dir=output_dir,
                model=model, optimizer=optimizer, scheduler=scheduler,
                device=device
            )
        else: print(f'\033[91mStarting from scratch\033[0m')
        
        # Collect all training configurations for logging
        train_config = model_config.copy()
        train_config.update({
            "benchmark": benchmark_name,
            "model": model_name,
            "seed": seed,
            "bsz": batch_size,
            "accumulation_steps": accumulation_steps,
            "lr": optimizer_config.get("lr"),
            "weight_decay": weight_decay,
            "grad_clip_norm": grad_clip_norm,
            "permuted": dataset_splits["train"].get("permuted"),
            "min_len": train_loader.dataset.min_len if hasattr(train_loader.dataset, "min_len") else None,
            "max_len": train_loader.dataset.max_len if hasattr(train_loader.dataset, "max_len") else None,
        })
        
        # Train the model
        successful = True
        try:
            train(
                epochs=epochs, train_steps=train_steps, benchmark_name=benchmark_name, model_name=model_name,
                model=model, optimizer=optimizer, scheduler=scheduler,
                loss_fn=loss_fn, acc_fn=acc_fn, data_fn=data_fn,
                train_config=train_config, mixed_precision=mixed_precision,
                output_dir=output_dir,
                train_loader=train_loader, val_loader=val_loader, test_loader=test_loader,
                local_log_path=local_log_path,
                wandb_logging=wandb_logging, wandb_entity=wandb_entity, wandb_project=wandb_project,
                wandb_metrics=wandb_metrics,
                grad_clip_norm=grad_clip_norm, accumulation_steps=accumulation_steps,
                checkpoint_freq=checkpoint_freq, val_freq=val_freq, info_freq=info_freq,
                device=device,
            )
        except KeyboardInterrupt:
            successful = False
            terminate = input('Terminate all experiments? (y/n): ').strip().lower()
            if terminate == 'y': break
        except Exception as e:
            successful = False
            traceback_str = traceback.format_exc()
            print(f'\033[91mExperiment [{i + 1}/{len(experiments)}] failed with error:\n{traceback_str}\033[0m\n')
        if successful:
            successful_count += 1
            print(f'\033[92mExperiment [{i + 1}/{len(experiments)}] completed successfully\033[0m\n')
        
        del general_config, model_config, optimizer_config, scheduler_config, train_config
    
    print(f'\033[1m{successful_count}/{len(experiments)} experiments completed successfully\033[0m')


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from einops import rearrange, repeat
from contextlib import nullcontext, contextmanager

@contextmanager
def no_param_grad(model):
    requires_grad_states = {}
    for name, param in model.named_parameters():
        requires_grad_states[name] = param.requires_grad
        param.requires_grad = False
    try:
        yield
    finally:
        for name, param in model.named_parameters():
            param.requires_grad = requires_grad_states[name]

class PCLayer(nn.Module):
    def __init__(self, f_module, b_module, f_energy_fn=F.mse_loss, b_energy_fn=F.mse_loss, device=torch.device('cpu')):
        super().__init__()
        self.f_module = f_module
        self.b_module = b_module
        self.f_energy_fn = f_energy_fn
        self.b_energy_fn = b_energy_fn
        self.device = device
        self.to(device)
    
    @torch.no_grad()
    def get_io_shapes(self, x=None, y=None):
        if x is not None: return self.f_module(x)
        elif y is not None: return self.b_module(y)
        else: raise ValueError("Either x or y must be provided")
    
    def forward_energy(self, x, y):
        """
        x -> p_y
        p_y, y -> loss
        """
        p_y = self.f_module(x)
        return self.f_energy_fn(p_y, y)

    def backward_energy(self, x, y):
        """
        p_x <- y
        p_x, x -> loss
        """
        p_x = self.b_module(y)
        return self.b_energy_fn(p_x, x)

class PCModel(nn.Module):
    def __init__(self, layers, max_its=1, min_energy=1e-1, energy_lr=1e-3, energy_optimizer_class=optim.SGD, energy_grad_clip_norm=None, iterative_train=False, state_init_dir="forward", device=torch.device('cpu')):
        super().__init__()
        assert isinstance(layers, (list, nn.ModuleList)), "Layers must be a list or torch.nn.ModuleList"
        for layer in layers: assert isinstance(layer, PCLayer), "All layers must be of type PCLayer"
        self.layers = nn.ModuleList(layers) if type(layers) is list else layers
        self.cached_x_shape = None
        self.cached_y_shape = None
        self.max_its = max_its
        self.min_energy = min_energy
        self.energy_lr = energy_lr
        self.energy_optimizer_class = energy_optimizer_class
        self.energy_grad_clip_norm = energy_grad_clip_norm
        self.iterative_train = iterative_train
        self.state_init_dir = state_init_dir
        self.device = device
        self.is_pc_model = True
        self.to(device)
    
    def _zero_param_grads(self):
        for layer in self.layers:
            for param in layer.parameters():
                if param.grad is not None:
                    param.grad.zero_()
    
    def _normalize_param_grads(self):
        for layer in self.layers:
            for param in layer.parameters():
                if param.grad is not None:
                    param.grad.div_(self.max_its)
    
    @torch.no_grad()
    def get_io_shapes(self, x=None, y=None):
        if x is not None:
            for layer in self.layers:
                x = layer.get_io_shapes(x, None)
            return x.shape
        elif y is not None:
            for layer in reversed(self.layers):
                y = layer.get_io_shapes(None, y)
            return y.shape
    
    @torch.no_grad()
    def update_io_shapes(self, x=None, y=None):
        if (x is not None and self.cached_x_shape == x.shape) or (y is not None and self.cached_y_shape == y.shape): return
        if x is not None:
            self.cached_x_shape = x.shape
            zero_x = torch.zeros_like(x, device=self.device)
            self.cached_y_shape = self.get_io_shapes(zero_x)
        elif y is not None:
            self.cached_y_shape = y.shape
            zero_y = torch.zeros_like(y, device=self.device)
            self.cached_x_shape = self.get_io_shapes(zero_y)
    
    @torch.no_grad()
    def forwad_state_init(self, x):
        state_tensors = []
        for layer in self.layers:
            state_tensors.append(x.clone().detach().requires_grad_(True))
            x = layer.f_module(x)
        state_tensors.append(x.clone().detach().requires_grad_(True))
        return state_tensors
    
    @torch.no_grad()
    def backward_state_init(self, y):
        state_tensors = []
        for layer in reversed(self.layers):
            state_tensors.append(y.clone().detach().requires_grad_(True))
            y = layer.b_module(y)
        state_tensors.append(y.clone().detach().requires_grad_(True))
        return state_tensors[::-1]

    def compute_model_energy(self, state_tensors):
        energy = torch.stack([
            layer.forward_energy(state_tensors[j], state_tensors[j+1])
            for j, layer in enumerate(self.layers)
        ]+[
            layer.backward_energy(state_tensors[j], state_tensors[j+1])
            for j, layer in enumerate(self.layers)
        ]).sum()
        return energy
    
    def train_forward(self, x, y, scaler=None):
        self.update_io_shapes(x, y)
        
        # Amortized state initialization
        if self.state_init_dir == "forward":
            state_tensors = self.forwad_state_init(x)
            state_tensors[-1] = y
        elif self.state_init_dir == "backward":
            state_tensors = self.backward_state_init(y)
            state_tensors[0] = x
        else:
            raise ValueError("State initialization direction must be either 'forward' or 'backward'")
        
        energy_optimizer = self.energy_optimizer_class(state_tensors[1:-1], lr=self.energy_lr)
        
        if self.iterative_train:
            # Iterative training:
            # One energy convergence step per parameter gradient calculation
            for i in range(self.max_its):
                energy_optimizer.zero_grad()
                
                energy = self.compute_model_energy(state_tensors)
                
                if scaler is not None: scaler.scale(energy).backward()
                else: energy.backward()
                if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                
                energy_optimizer.step()
                
                # if energy.item() < self.min_energy: break
            self._normalize_param_grads() # Normalize gradients by number of iterations
        else:
            # Non-iterative training:
            # Energy convergence before parameter gradient calculation
            for i in range(self.max_its):
                if i == self.max_its - 1: self._zero_param_grads() # Only calculate param grads on final iteration
                energy_optimizer.zero_grad()
                
                energy = self.compute_model_energy(state_tensors)
                
                if scaler is not None: scaler.scale(energy).backward()
                else: energy.backward()
                if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                
                energy_optimizer.step()
                
                # if energy.item() < self.min_energy: break
        return energy
    
    def forward(self, x=None, y=None):
        assert not (x is None and y is None), "Either x or y must be provided"
        
        self.update_io_shapes(x, y)

        with nullcontext() if self.training else no_param_grad(self):
            if x is not None:
                # Forward inference
                state_tensors = self.forwad_state_init(x)
                
                energy_optimizer = self.energy_optimizer_class(state_tensors[1:], lr=self.energy_lr) # Exclude input
                for i in range(self.max_its):
                    energy_optimizer.zero_grad()
                    
                    energy = self.compute_model_energy(state_tensors)
                    
                    energy.backward()
                    if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                    
                    energy_optimizer.step()
                    
                    if energy.item() < self.min_energy: break
                return state_tensors[-1]
            elif y is not None:
                # Backward inference
                state_tensors = self.backward_state_init(y)
                
                energy_optimizer = self.energy_optimizer_class(state_tensors[:-1], lr=self.energy_lr) # Exclude output
                for i in range(self.max_its):
                    energy_optimizer.zero_grad()
                    
                    energy = self.compute_model_energy(state_tensors)
                    
                    energy.backward()
                    if self.energy_grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(state_tensors, max_norm=self.energy_grad_clip_norm)
                    
                    energy_optimizer.step()
                    
                    if energy.item() < self.min_energy: break
                return state_tensors[0]

# from torchvision import datasets, transforms
# from torch.utils.data import DataLoader

# device = torch.device('cpu')

# transform = transforms.Compose([transforms.ToTensor(),])
# train_set = datasets.MNIST(root='./data', train=True, download=True, transform=transform)
# test_set = datasets.MNIST(root='./data', train=False, download=True, transform=transform)
# train_loader = DataLoader(train_set, batch_size=64, shuffle=True)
# test_loader = DataLoader(test_set, batch_size=64, shuffle=False)

# model = PCModel(
#     layers=[
#         PCLayer(nn.Sequential(nn.Linear(784, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 784, bias=False)), device=device),
#         PCLayer(nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), device=device),
#         # PCLayer(nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), device=device),
#         # PCLayer(nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), device=device),
#         # PCLayer(nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), nn.Sequential(nn.Linear(1024, 16, bias=False), nn.SiLU(), nn.Linear(16, 1024, bias=False)), device=device),
#         PCLayer(nn.Linear(1024, 10, bias=False), nn.Linear(10, 1024, bias=False), f_energy_fn=F.cross_entropy, b_energy_fn=F.mse_loss, device=device),
#     ],
#     max_its=3,
#     min_energy=1e-3,
#     energy_lr=1e-2,
#     energy_optimizer_class=optim.SGD,
#     device=device
# )

# param_optimizer = optim.AdamW(model.parameters(), lr=1e-3, weight_decay=0)

# def train():
#     model.train()
#     for batch_idx, (data, target) in enumerate(train_loader):
#         data = data.flatten(1).to(device)
#         target = target.to(device)
#         target = F.one_hot(target, num_classes=10).float()
#         param_optimizer.zero_grad()
#         train_energy = model.train_forward(data, target)
#         torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
#         param_optimizer.step()
#         if batch_idx % 100 == 0 and batch_idx > 0:
#             with no_param_grad(model):
#                 y = model(data)
#                 acc = (y.argmax(dim=1) == target.argmax(dim=1)).float().mean()
#                 loss = F.cross_entropy(y, target)
#                 print(f"Train epoch: {epoch}, batch: {batch_idx}", train_energy.item(), loss.item(), 100*acc.item())

# def test():
#     model.eval()
#     total_loss = 0
#     correct = 0
#     for data, target in test_loader:
#         data = data.flatten(1).to(device)
#         target = target.to(device)
#         y = model(data)
#         loss = F.cross_entropy(y, target)
#         total_loss += loss.item()
#         correct += (y.argmax(dim=1) == target).sum().item()

#     print("Test Loss:", total_loss / len(test_loader))
#     print("Test Accuracy:", 100 * correct / len(test_loader.dataset))

# epochs = 10
# for epoch in range(epochs):
#     train()
#     test()

# from .models.transformers import MultiheadAttention
# class PCTransformer(PCModel):
#     def __init__(self, emb_dim, input_dim, output_dim,
#                  n_layers=1, n_heads=1, mlp_dim=None,
#                  dropout=0.0, causal=True, use_embedding=True, weight_tying=False,
#                  mlp_bias=True, attention_bias=True,
#                  pos_encoding=None, pos_encoding_max_len=None,
#                  device="cpu"):
#         class PCTransformerLayer(nn.Module):
#             def __init__(self, emb_dim, n_heads, mlp_dim,
#                         dropout=0.0, causal=True, use_embedding=True,
#                         weight_tying=False, mlp_bias=True,
#                         attention_bias=True,
#                         pos_encoding=None, pos_encoding_max_len=None, device="cpu"):
#                 super().__init__()
#                 self.emb_dim = emb_dim
#                 self.n_heads = n_heads
#                 self.mlp_dim = mlp_dim
#                 self.dropout = dropout
#                 self.causal = causal
#                 self.use_embedding = use_embedding
#                 self.weight_tying = weight_tying
#                 self.mlp_bias = mlp_bias
#                 self.attention_bias = attention_bias
#                 self.pos_encoding = pos_encoding
#                 self.pos_encoding_max_len = pos_encoding_max_len
#                 self.attention = MultiheadAttention(emb_dim, n_heads, bias=attention_bias, batch_first=True)
#                 self.norm1 = nn.LayerNorm(emb_dim)
#                 self.dropout1 = nn.Dropout(dropout)
#                 self.norm2 = nn.LayerNorm(emb_dim)
#                 self.dropout2 = nn.Dropout(dropout)
#                 self.feedforward = nn.Sequential(
#                     nn.Linear(emb_dim, mlp_dim, bias=mlp_bias),
#                     nn.ReLU(),
#                     nn.Dropout(dropout),
#                     nn.Linear(mlp_dim, emb_dim, bias=mlp_bias)
#                 )
#                 if pos_encoding == "rope":
#                     self.rope = RotaryEmbedding(dim=emb_dim//(2*n_heads), use_xpos=False, cache_if_possible=False)
#                 elif pos_encoding == "xpos":
#                     self.rope = RotaryEmbedding(dim=emb_dim//(2*n_heads), use_xpos=True, cache_if_possible=False)
#                 elif pos_encoding == "abs":
#                     assert pos_encoding_max_len is not None, "pos_encoding_max_len must be provided for absolute positional encoding"
#                     self.pos_encoding_max_len = pos_encoding_max_len
#                 else: self.rope = None
#                 self.abs_pos_encoding = nn.Embedding(pos_encoding_max_len, emb_dim) if pos_encoding == "abs" else None
                
#                 self.to(device)
            
#             def forward(self, x):
#                 x = self.norm1(x)
#                 if self.abs_pos_encoding is not None:
#                     pos = torch.arange(x.size(1), device=x.device, dtype=torch.long).unsqueeze(0).expand(x.size(0), -1)
#                     x = x + self.abs_pos_encoding(pos)
#                 a_out, _ = self.attention(x, attn_mask=None, rope=self.rope if self.pos_encoding == "rope" else None)
#                 x = x + self.dropout1(a_out)
#                 x = self.norm2(x)
#                 ff_out = self.feedforward(x)
#                 x = x + self.dropout2(ff_out)
#                 return x

#         f_in_proj = nn.Linear(input_dim, emb_dim, bias=False, device=device)
#         b_in_proj = nn.Linear(emb_dim, input_dim, bias=False, device=device)
#         f_out_proj = nn.Linear(emb_dim, output_dim, bias=False, device=device)
#         b_out_proj = nn.Linear(output_dim, emb_dim, bias=False, device=device)
#         layers = [
#                 PCLayer(
#                     f_in_proj,
#                     b_in_proj,
#                     f_energy_fn=F.mse_loss,
#                     b_energy_fn=F.mse_loss,
#                     device=device
#                 )
#             ]+[
#             PCLayer(
#                 PCTransformerLayer(emb_dim, n_heads, mlp_dim,
#                                    dropout=dropout, causal=causal, use_embedding=use_embedding,
#                                    weight_tying=weight_tying, mlp_bias=mlp_bias,
#                                    attention_bias=attention_bias,
#                                    pos_encoding=pos_encoding, pos_encoding_max_len=pos_encoding_max_len, device=device),
#                 PCTransformerLayer(emb_dim, n_heads, mlp_dim,
#                                    dropout=dropout, causal=causal, use_embedding=use_embedding,
#                                    weight_tying=weight_tying, mlp_bias=mlp_bias,
#                                    attention_bias=attention_bias,
#                                    pos_encoding=pos_encoding, pos_encoding_max_len=pos_encoding_max_len, device=device),
#                 f_energy_fn=F.mse_loss,
#                 b_energy_fn=F.mse_loss,
#                 device=device
#             ) for _ in range(n_layers)
#         ]+[
#             PCLayer(
#                     nn.Sequential(
#                         nn.LayerNorm(emb_dim),
#                         f_out_proj
#                     ),
#                     nn.Sequential(
#                         b_out_proj,
#                         nn.LayerNorm(emb_dim),
#                     ),
#                     f_energy_fn=F.cross_entropy,
#                     b_energy_fn=F.mse_loss,
#                     device=device
#                 )
#         ]
        
#         nn.init.xavier_uniform_(f_in_proj.weight)
#         nn.init.xavier_uniform_(b_in_proj.weight)
#         if weight_tying:
#             f_out_proj.weight = f_in_proj.weight
#             b_out_proj.weight = b_in_proj.weight
#         else:
#             nn.init.xavier_uniform_(f_out_proj.weight)
#             nn.init.xavier_uniform_(b_out_proj.weight)
        
#         super().__init__(
#             layers=layers,
#             max_its=3,
#             min_energy=1e-3,
#             energy_lr=1e-2,
#             energy_optimizer_class=torch.optim.SGD,
#             energy_grad_clip_norm=None,
#             iterative_train=False,
#             state_init_dir="forward", 
#             device=device,
#         )

# def pc_train_epoch(epoch, train_loader, model, optimizer, loss_fn, acc_fn, data_fn=default_data_fn,
#                 scheduler=None, device="cpu",
#                 output_dir="", model_name=None, val_loader=None,
#                 wandb_logging=True, wandb_metrics=["acc", "loss"],
#                 grad_clip_norm=None, accumulation_steps=0,
#                 mixed_precision=False,
#                 checkpoint_freq=None, val_freq=None, info_freq=None):
#     # Default model name
#     if model_name is None: model_name = model.__class__.__name__
#     model.train()
#     train_loss = 0
#     train_acc = 0
#     iterable = tqdm(train_loader, desc=f"Train Epoch {epoch}", leave=False, bar_format='{desc}: [{n_fmt}/{total_fmt}] {percentage:.0f}%|{bar}| [{rate_fmt}] {postfix}')
#     scaler = GradScaler(device=device) if mixed_precision else None
#     optimizer.zero_grad()
#     for batch_idx, (data, target) in enumerate(iterable):
#         with autocast(device_type=device) if mixed_precision else nullcontext():
#             # Train pass
#             data = data.to(device)
#             target = target.to(device)
#             data, target = data_fn(data, target)
#             output = model.train_forward(data, target)
        
#         if (batch_idx + 1) % (accumulation_steps + 1) == 0:
#             if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
#             optimizer.step() if not mixed_precision else scaler.step(optimizer)
#             optimizer.zero_grad()
#             if scheduler: scheduler.step()
#             if mixed_precision: scaler.update()
            
#             model.eval()
#             output = model(data)
#             model.train()
            
#             # Accuracy
#             with torch.no_grad():
#                 accuracy = acc_fn(output.clone(), target.clone())
#                 train_acc += accuracy
            
#             # Loss
#             loss = loss_fn(output, target)
#             train_loss += loss.item()
            
#             # WandB logging
#             if wandb_logging:
#                 log_data = {}
#                 if "acc" in wandb_metrics: log_data["train/acc"] = accuracy
#                 if "loss" in wandb_metrics: log_data["train/loss"] = loss.item()
#                 if "ppl" in wandb_metrics: log_data["train/ppl"] = math.exp(loss.item())
#                 if "lr" in wandb_metrics: log_data["misc/lr"] = scheduler.get_last_lr()[0]
#                 if "seq_len" in wandb_metrics: log_data["misc/seq_len"] = train_loader.dataset.len
#                 wandb.log(log_data)
        
#         # Post info
#         if info_freq and batch_idx % info_freq == 0 and batch_idx != 0:
#             tqdm.write(f'Train Epoch {epoch}: [{batch_idx}/{len(train_loader)}] LR: {scheduler.get_last_lr()[0]:.1e}, Loss: {loss.item():.4f}, Acc: {accuracy:.2f}%')
        
#         # Checkpoint
#         if checkpoint_freq and batch_idx % checkpoint_freq == 0 and batch_idx != 0:
#             checkpoint(model_name, output_dir, model, optimizer, scheduler)
        
#         # Validation
#         if val_freq and batch_idx % val_freq == 0 and batch_idx != 0:
#             if val_loader: val_epoch(model, val_loader, loss_fn, acc_fn, device=device, wandb_logging=wandb_logging, wandb_metrics=wandb_metrics)
#             model.train()
    
#     # Account for last accumulated batch
#     if (batch_idx + 1) % (accumulation_steps + 1) != 0:
#         if grad_clip_norm is not None: torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=grad_clip_norm)
#         optimizer.step() if not mixed_precision else scaler.step(optimizer)
#         optimizer.zero_grad()
#         if scheduler: scheduler.step()
#         if mixed_precision: scaler.update()
        
#         model.eval()
#         output = model(data)
#         model.train()
        
#         # Accuracy
#         with torch.no_grad():
#             accuracy = acc_fn(output.clone(), target.clone())
#             train_acc += accuracy
        
#         # Loss
#         loss = loss_fn(output, target)
#         train_loss += loss.item()
        
#         # WandB logging
#         if wandb_logging:
#             log_data = {}
#             if "acc" in wandb_metrics: log_data["train/acc"] = accuracy
#             if "loss" in wandb_metrics: log_data["train/loss"] = loss.item()
#             if "ppl" in wandb_metrics: log_data["train/ppl"] = math.exp(loss.item())
#             if "lr" in wandb_metrics: log_data["misc/lr"] = scheduler.get_last_lr()[0]
#             if "seq_len" in wandb_metrics: log_data["misc/seq_len"] = train_loader.dataset.len
#             wandb.log(log_data)
    
#     # Step sequence length if applicable
#     if hasattr(train_loader.dataset, "step"):
#         train_loader.dataset.step()
    
#     train_loss /= len(train_loader) / (accumulation_steps + 1)
#     train_acc /= len(train_loader) / (accumulation_steps + 1)
    
#     return train_loss, train_acc